import json
import os
import re
from functools import wraps

import requests
from dotenv import load_dotenv
from flask import Flask, request, Response

# 加载环境变量
load_dotenv()

app = Flask(__name__)

# 目标API地址
TARGET_API_BASE = "https://api.bianxieai.com/v1"


def replace_escaped_newlines(text):
    return re.sub(r'\nin (.*?)\n', r'\n\1\n', text)


def error_handler(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            return Response(
                str(e),
                status=500,
                mimetype='application/json'
            )

    return decorated_function


@app.route('/<path:path>', methods=['GET', 'POST', 'PUT', 'DELETE'])
@error_handler
def proxy(path):
    # 构建目标URL
    target_url = f"{TARGET_API_BASE}/{path}"

    # 获取原始请求的headers
    headers = dict(request.headers)

    # 移除一些不需要转发的headers
    headers.pop('Host', None)
    headers.pop('Content-Length', None)

    # 获取请求数据
    data = request.get_data()

    # 发送请求到目标API
    resp = requests.request(
        method=request.method,
        url=target_url,
        headers=headers,
        data=data,
        stream=True
    )

    # 处理响应内容
    try:
        res = json.loads(resp.content.decode())
        if 'choices' in res and len(res['choices']) > 0 and 'message' in res['choices'][0]:
            res_content = res['choices'][0]['message']['content']
            res_content = replace_escaped_newlines(res_content)
            res['choices'][0]['message']['content'] = res_content

            # 将修改后的内容转换回Response对象
            excluded_headers = ['content-encoding', 'content-length', 'transfer-encoding', 'connection']
            headers = [(name, value) for (name, value) in resp.raw.headers.items()
                       if name.lower() not in excluded_headers]

            return Response(
                json.dumps(res),
                resp.status_code,
                headers
            )
    except (json.JSONDecodeError, KeyError, AttributeError):
        pass

    # 如果处理失败或不需要处理，返回原始响应
    excluded_headers = ['content-encoding', 'content-length', 'transfer-encoding', 'connection']
    headers = [(name, value) for (name, value) in resp.raw.headers.items()
               if name.lower() not in excluded_headers]

    return Response(
        resp.content,
        resp.status_code,
        headers
    )


@app.route('/health')
def health_check():
    return {"status": "healthy"}, 200


if __name__ == '__main__':
    # 设置端口，默认为5000
    port = int(os.getenv('PORT', 5100))
    # 设置是否开启调试模式
    debug = os.getenv('DEBUG', 'False').lower() == 'true'

    app.run(
        host='0.0.0.0',
        port=port,
        debug=debug
    )